# Visualisation code written by M. Crucifix 2014, 
# used for supplementary material in 
# Supplementary Material to Article : 
# P. A. Araya Melo, M. Crucifix, N. Bounceur, 
# Global sensitivity analysis of Indian Monsoon during the Pleistocene
# http://www.clim-past-discuss.net/10/1609/2014/cpd-10-1609-2014.html
#
# using the enthought traits library

# http://code.enthought.com/support/
# ETS projects are licensed under Enthought's BSD-style license.
# On Mac OS, the visualisation requires to use the Qt backends 
# take note of the following : 
# However, you should note that specific usage of the optional Qt backends in ETS,
# instead of the default WX backends, requires that your application also use Qt and PyQt, 
# which are NOT BSD licensed, and thus may impact your distribution.
# Please see TrollTech, Riverbank Computing and the PySide project for more information. 


# for all other code (supplied explitictly here), the MIT Licence applies

from traits.api import HasTraits, Range, Instance, Enum, List, Float, \
                    on_trait_change, Bool
from traitsui.api import View, Item, HGroup, RangeEditor
from tvtk.pyface.scene_editor import SceneEditor
from mayavi.tools.mlab_scene_model import \
                    MlabSceneModel
from tvtk.pyface.scene_model import SceneModel
from mayavi.core.ui.mayavi_scene import MayaviScene
from mayavi.core.ui.engine_view import EngineView
from mayavi.core.api import PipelineBase, Source
from mayavi import mlab
import numpy as np
from rpy2 import robjects

def memoize(f):
    """ Memoization decorator for functions taking one or more arguments. """
    class memodict(dict):
        def __init__(self, f):
            self.f = f
        def __call__(self, *args):
            return self[args]
        def __missing__(self, key):
            ret = self[key] = self.f(*key)
            return ret
    return memodict(f)

@memoize
def load_data(r,ice,co2,orog,l_ecosw,l_esinw,l_eps,l_co2,l_ice,nugget):
  """ load_data by calling the r backend """
  x,y,z = np.mgrid[0:1:11j, 0:1:11j, 0:1:11j]
  s = np.array( r.P(orog,ice/10.,co2/10.,l_ecosw,l_esinw,l_eps,l_co2,l_ice,nugget))
  vmin = np.min(s)
  vmax = np.max(s)
  return x,y,z,s,vmin,vmax
    

def calc_dist(r,ice,co2):
    ice_dist  = ( (ice/10. - np.array( r.X.rx(True,'ice') ) )/ r.lambd.rx2('theta').rx('ice') )**2 
    co2_dist  = ( (co2/10. - np.array( r.X.rx(True,'co2') ) )/ r.lambd.rx2('theta').rx('co2') )**2 
    dists=np.exp(- ice_dist - co2_dist)
    return(dists)

class Visualization(HasTraits):
    scene_x = Instance(MlabSceneModel, ())
    scene_y = Instance(MlabSceneModel, ())
    scene_z = Instance(MlabSceneModel, ())
    scene3d = Instance(MlabSceneModel, ())

    src = Instance(Source)

    tnames = List([])
    varnames = List([])
    masknames = List([])
    typenames = List([])

    vmin      = Float
    vmax      = Float
    contour   = Float
    co2    = Range(0,10,5, auto_set=False, enter_set=True)
    ice    = Range(0,10,5, auto_set=False, enter_set=True)
    l_ecosw = Range(0.10,2.20,0.5, auto_set=True, enter_set=True, format='%.2f')
    l_esinw = Range(0.10,2.00,0.5, auto_set=True, enter_set=True, format='%.1f')
    l_eps = Range(0.10,2.00,0.5, auto_set=True, enter_set=True, format='%.1f')
    l_co2 = Range(0.10,2.00,0.5, auto_set=True, enter_set=True, format='%.1f')
    l_ice = Range(0.10,2.20,0.5, auto_set=True, enter_set=True, format='%.1f')
    nugget = Range(0,0.20,0, auto_set=False, enter_set=True, format='%.2f')
    showdesign    = Bool()
    orog          = Bool()
    mask          = Bool()

    def __init__(self,r,design=None):
        # Do not forget to call the parent's __init__
        HasTraits.__init__(self)
        self.r=r

        x,y,z,s,self.vmin,self.vmax = load_data(self.r, self.ice, self.co2,self.orog,
            self.l_ecosw,self.l_esinw, self.l_eps,self.l_co2, self.l_ice, self.nugget)
        self.source = self.scene3d.mlab.pipeline.scalar_field(x,y,z,s)
        self.ipws = {}
        for s in ('x','y','z'):
          self.ipws[s]=self.scene3d.mlab.pipeline.image_plane_widget(self.source, 
          plane_orientation='%s_axes' % s, vmin=self.vmin, vmax=self.vmax)
          self.ipws[s].ipw.reslice_interpolate = 'nearest'
          dists       = calc_dist(self.r,self.ice, self.co2)
          self.points = self.scene3d.mlab.points3d(
            r.X.rx(True,'ecosw'), r.X.rx(True,'esinw'), r.X.rx(True,'eps'), dists,
             scale_mode='scalar', color=(1,1,1), scale_factor=0.15)
          self.points.visible = self.showdesign
        
    @on_trait_change('scene3d.activated')
    def display_scene3d(self):
       self.surf = self.scene3d.mlab.pipeline.iso_surface(
                 self.source, contours=([-32.,]),
                 opacity=0.7, color=(0.2,0.2,0.2), colormap='gist_ncar')
       self.scene3d.mlab.axes(self.source, xlabel='ecosw', ylabel='esinw', zlabel='eps')

       lut = self.ipws['x'].module_manager.scalar_lut_manager.lut.table.to_array()
       # set last color to transparent
       lut[255, -1] = 0
       self.ipws['x'].module_manager.scalar_lut_manager.lut.table = lut
       self.scene3d.mlab.colorbar()

    
    @on_trait_change('ice,co2,l_ecosw,l_esinw,l_eps,l_co2,l_ice,nugget,orog')
    def update_plot(self):
        self.l_ecosw = np.round(self.l_ecosw, decimals=1)
        self.l_esinw = np.round(self.l_esinw, decimals=1)
        self.l_eps = np.round(self.l_eps, decimals=1)
        self.l_co2 = np.round(self.l_co2, decimals=1)
        self.l_ice = np.round(self.l_ice, decimals=1)
        self.nugget = np.round(self.nugget, decimals=2)
        x,y,z,s,self.vmin,self.vmax = load_data(self.r,self.ice, self.co2,self.orog,
            self.l_ecosw,self.l_esinw, self.l_eps,self.l_co2, self.l_ice, self.nugget)
        dists  = calc_dist(self.r,self.ice, self.co2)
        self.surf.mlab_source.set(x=x,y=y,z=z,scalars=s)
        self.points.mlab_source.set(scalars=dists)
        self.ipws['z'].module_manager.scalar_lut_manager.lut.range = (self.vmin,self.vmax)


    @on_trait_change('contour')
    def udpdate_contour(self):
       self.surf.contour.contours = [self.contour,]

    @on_trait_change('showdesign')
    def update_showdesign(self):
       self.points.visible=self.showdesign

    # the layout of the dialog created
    view = View(
                Item('scene3d', editor=SceneEditor(scene_class=MayaviScene),
                    height=400, width=400, show_label=True),
                HGroup(
                       'orog','ice','co2','showdesign'
                    ),
                HGroup(
                       'l_ecosw', 'l_esinw', 'l_eps'
                      ),
                HGroup(
                       'l_co2', 'l_ice', 'nugget'
                      ),
                Item('contour', 
                      editor=RangeEditor(low_name='vmin', high_name='vmax', 
                        mode='auto',  format = '%.1f')
                    ),
 
                )



def main():
  import os
  # see the above note about licence 
  # on MAC OS, you may have to enter 
  # export ETS_TOOLKIT='qt4' on the shell befor execution
  os.environ['ETS_TOOLKIT'] = 'qt4'
  from rpy2.robjects import r

  r("source('read_orog.R')")
  a = Visualization(r)
  a.configure_traits()

import sys
if __name__ == "__main__":
    sys.exit(main())
